import sys
sys.path.append('.')

import pytest
from lib.data import DataManager
from lib.utils import decode_distribution

data_det = DataManager.load('assets/det.json')
data_mc  = DataManager.load('assets/mc.json')
data_ml  = DataManager.load('assets/ml.json')

record_det1 = {
    "info": {
            "uuid": "59fe019ce0722fad5ae2b6fcba9b68a9",
            "image_path": "MP2112712A_HJB169-04-F3_UBM_9_1_1.jpg"
        },
        "instances": [
            {
                "uuid": "73db27eb3785d6523b6dfa32b1dcf693",
                "shape_type": "polygon",
                "points": "298+344,298+406,368+406,368+344,298+406.",
                "score": 0.06946061,
                "classname": "BD"
            },
            {
                "uuid": "73db27eb3785d6523b6dfa32b1dcf693",
                "shape_type": "polygon",
                "points": "298+644,298+406,368+406,368+644,298+406.",
                "score": 0.7825,
                "classname": "FM"
            }
        ]
}
record_det2 = {
    "info": {
            "uuid": "59fe019ce0722fad5ae2b6fcba9b68a9",
            "image_path": "MP2112712A_HJB169-04-F3_UBM_9_1_1.jpg"
        },
        "instances": []
}
record_mc = {
    "info": {
        "uuid": "bff6e80a1b0ad51f60ab6159a64d411f",
        "image_path": "X26477-01B4_X_74_Y_75_215_1.jpg"
    },
    "distribution": "1.733311e-01,0.0,0.0,0.0,0.0,0.0,0.0,0.0,8.266689e-01,0.0,0.0,0.0,0.0,0.0"
}
record_ml = {
    "info": {
        "uuid": "bff6e80a1b0ad51f60ab6159a64d411f",
        "image_path": "X26477-01B4_X_74_Y_75_215_1.jpg"
    },
    "scores": "0.0,0.0,0.0,4.889707e-02,4.368331e-02,9.424161e-02,0.0,8.266689e-01,6.326669e-02,0.0,0.0,0.0,0.0"
}



@pytest.mark.parametrize('data', [data_det, data_mc, data_ml])
def test_dataset_format(data):
    class_list = data.class_list
    # check class list
    assert isinstance(class_list, list)
    assert all([isinstance(classname, str) for classname in class_list])

    # check record
    record_list = data.record_list
    assert isinstance(record_list, list)

    ## check record
    task_type = data.task_type
    assert task_type in ['multiclass', 'multilabel', 'detection']

    for record in data:
        assert 'info' in record
        assert 'uuid' in record['info']
        assert 'image_path' in record['info']

        if task_type == 'multiclass':
            'distribution' in record
        elif task_type == 'multilabel':
            'scores' in record
        else:
            'instances' in record
            # check instances
            for inst in record['instances']:
                assert 'uuid' in inst
                assert 'shape_type' in inst
                assert 'points' in inst
                assert 'score' in inst
                assert 'classname' in inst


def test_detection_info_extraction():
    data = data_det
    instance = record_det1['instances'][0]

    ## classid
    classid = data.get_detection_classid_from_instance(instance)
    assert classid == 2
    classid_list = data.get_detection_classid_list_from_record(record_det1)
    assert len(classid_list) == 2 and classid_list[0] == 2 and classid_list[1] == 5
    classid_list = data.get_detection_classid_list_from_record(record_det2)
    assert len(classid_list) == 0

    ## classname
    classname = data.get_detection_classname_from_instance(instance)
    assert classname == 'BD'
    classname_list = data.get_detection_classname_list_from_record(record_det1)
    assert len(classname_list) == 2 and classname_list[0] == 'BD' and classname_list[1] == 'FM'
    classname_list = data.get_detection_classname_list_from_record(record_det2)
    assert len(classname_list) == 0

    ## score
    score = data.get_detection_score_from_instance(instance)
    assert score == .06946061
    scores = data.get_detection_score_list_from_record(record_det1)
    assert len(scores) == 2 and scores[0] == .06946061 and scores[1] == .7825
    scores = data.get_detection_score_list_from_record(record_det2)
    assert len(scores) == 0

    ## shape
    points = data.get_detection_shape_from_instance(instance)
    assert len(points) == 5 and points[0][1] == 344
    pt_list = data.get_detection_shape_list_from_record(record_det1)
    assert len(pt_list) == 2 and len(pt_list[0]) == 5 and pt_list[0][0][1] == 344
    xyxy = data.get_detection_xyxy_from_instance(instance)
    assert tuple(xyxy) == (298.0, 344.0, 368.0, 406.0)
    xyxys = data.get_detection_xyxy_list_from_record(record_det1)
    assert len(xyxys) == 2 and tuple(xyxys[1]) == (298.0, 406.0, 368.0, 644.0)
    xywh = data.get_detection_xywh_from_instance(instance)
    assert tuple(xywh) == (298.0, 344.0, 70.0, 62.0)
    xywhs = data.get_detection_xywh_list_from_record(record_det1)
    assert len(xywhs) == 2 and tuple(xywhs[1]) == (298.0, 406.0, 70.0, 238.0)


def test_multiclass_info_extraction():
    data = data_mc
    record = record_mc
    distribution = decode_distribution(record['distribution'])

    ## classid
    classid = data.get_multiclass_classid_from_distribution(distribution)
    assert classid == 8
    classid = data.get_multiclass_classid_from_record(record)
    assert classid == 8

    ## classname
    classname = data.get_multiclass_classname_from_distribution(distribution)
    assert classname == 'BS_White'
    classname = data.get_multiclass_classname_from_record(record)
    assert classname == 'BS_White'

    ## score
    distribution = data.get_multiclass_distribution_from_record(record)
    assert tuple(distribution) == (0.1733311, 0, 0, 0, 0, 0, 0, 0, 0.8266689, 0, 0, 0, 0, 0)
    score = data.get_multiclass_score_from_distribution(distribution)
    assert score == .8266689
    score = data.get_multiclass_score_from_record(record)
    assert score == .8266689



def test_multiclass_info_extraction():
    data = data_ml
    record = record_ml

    ## scores
    scores = data.get_multilabel_scores_from_record(record)
    assert tuple(scores) == (0, 0, 0, 0.04889707, 0.04368331, 0.09424161, 0, 0.8266689, 0.06326669, 0, 0, 0, 0)



#
# Class for evaluation (classification/detection/segmentation)
#

import numpy as np
import pandas as pd

from pathlib import Path
from tqdm import tqdm

from .base import EvaluatorBase

from ..processing import compute_box_box_iou
from ..utils import dir_check
from ..processing import cv_rgb_imread
from ..utils import encode_path
from ..processing import boxes_painter
from ..processing import cv_rgb_imwrite

from ..utils import eval_map


class DetectionEvaluator(EvaluatorBase):
    TYPE = 'Detection'

    def _get_pred_result_dict(self, dm_pred):
        # collect pred data for ap computation
        pred_result_dict = dict()
        for rec in dm_pred:
            # collect detection results
            bboxes = np.reshape(dm_pred.get_detection_xyxy_list_from_record(rec),  [-1,4])
            scores = np.reshape(dm_pred.get_detection_score_list_from_record(rec), [-1,1])
            labels = np.reshape(dm_pred.get_detection_classid_list_from_record(rec), [-1])
            det = np.hstack([bboxes, scores])
            det_res = [det[labels==classid,:] for classid in range(len(self.classnames))]
            pred_result_dict[rec['info']['uuid']] = det_res

        return pred_result_dict

    def _get_gt_result_dict(self, dm_gt):
        # collect gt data for ap computation
        gt_result_dict = dict()
        for rec in dm_gt:
            # collect annotation results
            ann = dict(
                bboxes = np.reshape(dm_gt.get_detection_xyxy_list_from_record(rec), [-1,4]),
                labels = np.reshape(dm_gt.get_detection_classid_list_from_record(rec), [-1])
            )
            gt_result_dict[rec['info']['uuid']] = ann

        return gt_result_dict

    def eval_ap(self, iou_threshold=.1, verbose=True):
        """
        :iou_threshold: the minimal iou that a TP prediction is required when overlapped with a gt bbox
        """
        # compute ap
        _, ap_results = eval_map(
            det_results=self.pred_data_list,
            annotations=self.gt_data_list,
            classnames=self.classnames,
            iou_thr=iou_threshold,
        )

        # print evaluation
        if verbose:
            ap_dict = {classname:x['ap'] for classname, x in zip(self.classnames, ap_results)}
            df = pd.DataFrame.from_dict(ap_dict, orient='index', columns=['AP'])
            df_notnull = df[df['AP'].notnull()]  # remove NaN data (No samples of this class in gt)
            print(df_notnull)

        wmAP = 0
        data_num = 0
        for ap_result in ap_results:
            wmAP += ap_result['num_gts'] * ap_result['ap']
            data_num += ap_result['num_gts']
        if data_num == 0:
            wmAP = 0
        else:
            wmAP /= data_num

        if verbose:
            print(f" wmAP: {wmAP}")

        eval_result = {
            'AP': ap_results,
            'wmAP': wmAP
        }

        self.ap_result = eval_result
        return eval_result

    def _set_threshold(self, policy='inflection', value=None):
        """
        Args:
            policy (str, optional): could be one of 'inflection', 'recall', 'precision' or 'manual'. Defaults to 'inflection'.
            value (optional): 
                under 'inflection' policy: value is insignificant
                under 'recall' or 'precision': value should be the minimum float we can accept
                under 'manual': value can be a list or dict who has the same length or keys with classnames. 
                . Defaults to None.
        """
        assert policy in ['inflection', 'recall', 'precision', 'manual'], f"policy {policy} should in ['inflection', 'recall', 'precision', 'manual']"
        threshold_dict = dict()

        if policy != 'manual':
            if not hasattr(self, 'ap_result'):
                print("@@ Warning: You did not run eval_ap manually, and we have to run it with default args.")
                self.eval_ap(verbose=False)
            assert hasattr(self, 'ap_result'), f"inflection threshold need self.ap_result, and we can not find it."

        if policy == 'inflection':
            ap_res = self.ap_result['AP']
            for key, ap_data in ap_res.items():
                recall = ap_data['recall']
                precision = ap_data['precision']
                score = ap_data['score']
                area = np.array(recall) * np.array(precision)
                if len(area) > 0 and area.max() > 0:
                    threshold_dict[key] = score[np.argmax(area)]
                else:
                    threshold_dict[key] = 1

        elif policy == 'recall':
            assert 0 <= value <= 1, f"bad value {value} for minimum recall"
            ap_res = self.ap_result['AP']
            for key, ap_data in ap_res.items():
                recall = ap_data['recall']
                mark = recall >= value
                score = ap_data['score']

                if mark.max() > 0:
                    threshold_dict[key] = score[np.argmax(mark)]
                else:
                    threshold_dict[key] = score[-1]

        elif policy == 'precision':
            assert 0 <= value <= 1, f"bad value {value} for minimum precision"
            ap_res = self.ap_result['AP']
            for key, ap_data in ap_res.items():
                precision = ap_data['precision']
                mark = precision >= value
                score = ap_data['score']

                if mark.max() > 0:
                    threshold_dict[key] = score[len(mark) - np.argmax(mark[::-1]) - 1]
                else:
                    threshold_dict[key] = score[0]
        elif policy == 'manual':
            if isinstance(value, dict):
                threshold_dict = value
            else:
                assert isinstance(value, list), f"under manual threshold policy, value should be dict or list. {value}"
                assert len(value) == len(self.classnames), f"value list is not as long as classnames {len(value)} != {len(self.classnames)}"
                for index in range(len(self.classnames)):
                    threshold_dict[self.classnames[index]] = value[index]

        else:
            raise RuntimeError(f"unknown policy {policy}")

        return threshold_dict

    def eval_judgment(self, iou_threshold=.1, verbose=True):
        """
        :iou_threshold: the minimal iou that a TP prediction is required when overlapped with a gt bbox
        """
        if not hasattr(self, 'threshold_dict'):
            print("@@ Warning: You did not run set_threshold manually, and we have to run it with default args.")
            self.set_threshold()
        assert hasattr(self, 'threshold_dict'), f"eval_judgment failed for threshold_dict not found."

        judge_res_list = list()
        for gt_data, pred_data, data_path in zip(self.gt_data_list, self.pred_data_list, self.data_path_list):
            judge_result = self.data_judge(gt_data, pred_data, self.threshold_dict, iou_threshold)
            judge_res_list.append([judge_result, data_path])
        self.judge_res_list = judge_res_list

    @staticmethod
    def data_judge(gt_data, pred_data, threshold_dict, iou_threshold):
        judge_result = [[] for _ in range(len(threshold_dict))]
        gt_bbox_dict = dict()
        for index, label in enumerate(gt_data['labels'].tolist()):
            if label not in gt_bbox_dict:
                gt_bbox_dict[label] = list()
            bbox_info = {
                'type': 'gt',
                'bbox': gt_data['bboxes'][index].tolist(),
                'hit': False
            }
            gt_bbox_dict[label].append(bbox_info)
            judge_result[label].append(bbox_info)
        for index, pred_boxes in enumerate(pred_data):
            for bbox_score in pred_boxes:
                bbox = bbox_score[:4].tolist()
                score = bbox_score[-1]

                if score < threshold_dict[index]:
                    continue
                else:
                    hit = False
                    if index in gt_bbox_dict:
                        for gt_bbox_info in gt_bbox_dict[index]:
                            gt_bbox = gt_bbox_info['bbox']
                            iou = compute_box_box_iou(gt_bbox, bbox)
                            if iou >= iou_threshold:
                                hit = True
                                gt_bbox_info['hit'] = True

                    bbox_info = {
                            'score': score,
                            'type': 'pred',
                            'bbox': bbox,
                            'hit': hit
                        }
                    judge_result[index].append(bbox_info)
        return judge_result

    def get_confusion_matrix(self):
        if not hasattr(self, 'judge_res_list'):
            print("@@ Warning: You did not run eval_judgment manually, and we have to run it with default args.")
            self.eval_judgment()
        assert hasattr(self, 'judge_res_list'), f"get_confusion_matrix failed for judge_res_list not found."
        cm_dict = {key:dict(TP=0, TN=0, FN=0, FP=0) for key in self.classnames}

        # y_true = np.zeros([len(self.judge_res_list), len(self.classnames)])
        # y_pred = np.zeros([len(self.judge_res_list), len(self.classnames)])

        for data_index, judge_res in enumerate(self.judge_res_list):
            for index, class_bboxes in enumerate(judge_res[0]):
                pred_label = False
                gt_label = False
                for bbox_info in class_bboxes:
                    if bbox_info['type'] == 'gt':
                        gt_label = True
                    elif bbox_info['type'] == 'pred':
                        pred_label = True
                    else:
                        raise ValueError(f"unknown type value {bbox_info['type']}")
                # if gt_label:
                #     y_true[data_index][index] = 1
                # if pred_label:
                #     y_pred[data_index][index] = 1
                
                if pred_label == gt_label == False:
                    cm_dict[self.classnames[index]]['TN'] += 1
                elif pred_label == gt_label == True:
                    cm_dict[self.classnames[index]]['TP'] += 1
                elif pred_label == True and gt_label == False:
                    cm_dict[self.classnames[index]]['FP'] += 1
                elif pred_label == False and gt_label == True:
                    cm_dict[self.classnames[index]]['FN'] += 1
                else:
                    raise ValueError(f"unknown pred_label {pred_label} and gt_label {gt_label} value.")

        for key, cm_info in cm_dict.items():
            assert cm_info['TP'] + cm_info['TN'] + cm_info['FP'] + cm_info['FN'] == len(self.judge_res_list), f"cm_info {cm_info} sum error."
        
        # cm_list = multilabel_confusion_matrix(y_true=y_true, y_pred=y_pred)
        
        cm_pd = pd.DataFrame(cm_dict)
        
        return cm_pd

    def dump_failure_case(self, data_root, target_dir):
        if not hasattr(self, 'judge_res_list'):
            print("@@ Warning: You did not run eval_judgment manually, and we have to run it with default args.")
            self.eval_judgment()
        assert hasattr(self, 'judge_res_list'), f"get_confusion_matrix failed for judge_res_list not found."

        dir_check(target_dir)
        target_dir = Path(target_dir)

        for judge_res in tqdm(self.judge_res_list):
            mode_list = list()

            box_list = list()
            label_list = list()
            score_list = list()
            color_list = list()

            draw = False

            for index, class_bboxes in enumerate(judge_res[0]):
                data_mode = set()
                label = self.classnames[index]
                for bbox_info in class_bboxes:
                    box = bbox_info['bbox']
                    if bbox_info['type'] == 'gt':
                        score = 1
                        if bbox_info['hit'] == False:
                            draw = True
                            data_mode.add('miss')
                            color = [255, 0, 0]
                        else:
                            color = [0, 255, 0]
                    elif bbox_info['type'] == 'pred':
                        score = bbox_info['score']
                        if bbox_info['hit'] == False:
                            draw = True
                            data_mode.add('false_alarm')
                            color = [255, 255, 0]
                        else:
                            color = [0, 0, 255]
                    else:
                        raise ValueError(f"unknown type value {bbox_info['type']}")
                    box_list.append(box)
                    label_list.append(label)
                    score_list.append(score)
                    color_list.append(color)

                mode_list.append(data_mode)
            if draw:
                ori_img_path = Path(data_root) / judge_res[1]
                if not ori_img_path.exists():
                    print(f"Ori image path {ori_img_path} not found.")
                    continue
                else:
                    try:
                        image = cv_rgb_imread(ori_img_path)
                    except Exception as e:
                        print(f"image load failed! {ori_img_path}")
                        print(f"error message: {e}")
                        continue
                painted_image = boxes_painter(image, box_list=box_list, label_list=label_list, score_list=score_list, color_list=color_list)

            for index, data_mode in enumerate(mode_list):
                label = self.classnames[index]
                for mode in data_mode:
                    save_path = target_dir / mode / label / encode_path(ori_img_path)
                    cv_rgb_imwrite(painted_image, save_path)

    eval = eval_ap
    evaluate = eval_ap
    evaluation = eval_ap





def url_validity_check(url, verbose=False):
    from urllib import request
    try:
        url = url.replace(' ', '%20')
        with request.urlopen(url) as file:
            if verbose:
                print(file.status)
                print(file.reason)
            return True
    except Exception as e:
        if verbose:
            print(e)
        return False



import copy

from .saver import DataManagerSaver
from ..utils import encode_distribution


class DataManagerClassManipulator(DataManagerSaver):
    def map_class(self, class_mapping_dict):
        """
        Manipulate the dataset classes by remapping current class_list to a new one

        Args:
            class_mapping_dict: a dictionary, which takes form of {old_classname: new_classname},
            if old_classname not found in class_mapping_dict, it remains the same

        Return:
            a new dataset with new class_list, and the class order is sorted
        """
        assert set(class_mapping_dict.keys()).issubset(self.class_list), \
            "Unknown classname (key) found: ".format(class_mapping_dict.keys())
        # 0) preparation
        oldclass2newclass = dict()
        for classname in self.class_list:
            if classname in class_mapping_dict:
                oldclass2newclass[classname] = class_mapping_dict[classname]
            else:
                oldclass2newclass[classname] = classname

        # 1) get new class_list
        new_class_list = sorted(list(set(oldclass2newclass.values())))
        new_class2label = {classname:classid for classid, classname in enumerate(new_class_list)}

        # 2) get new record_list
        data_task_type = self.task_type
        new_record_list = list()
        for record in self.record_list:
            new_record = copy.deepcopy(record)
            if data_task_type == 'detection':
                for inst in new_record['instances']:
                    old_classname = self.get_detection_classname_from_instance(inst)
                    new_classname = oldclass2newclass[old_classname]
                    inst['classname'] = new_classname
            elif data_task_type == 'multiclass':
                old_distribution = self.get_multiclass_distribution_from_record(record)
                new_distribution = [0] * len(new_class_list)
                for old_classid, old_classname in enumerate(self.class_list):
                    old_score = old_distribution[old_classid]
                    new_classname = oldclass2newclass[old_classname]
                    new_classid = new_class2label[new_classname]
                    new_distribution[new_classid] += old_score
                new_record['distribution'] = encode_distribution(new_distribution)
            else:
                assert data_task_type == 'multilabel'
                old_scores = self.get_multilabel_scores_from_record(record)
                new_scores = [0] * len(new_class_list)
                for old_classid, old_classname in enumerate(self.class_list):
                    old_score = old_scores[old_classid]
                    new_classname = oldclass2newclass[old_classname]
                    new_classid = new_class2label[new_classname]
                    new_scores[new_classid] = max(old_score, new_scores[new_classid])
                new_record['scores'] = encode_distribution(new_scores)

            new_record_list.append(new_record)

        assert len(new_record_list) == len(self.record_list)

        return type(self)(record_list=new_record_list, class_list=new_class_list)



# -*- coding: utf-8 -*-
# @Author: Zhai Menghua
# @Date:   2020-07-17 11:01:04
# @Last Modified by:   Zhai Menghua
# @Last Modified time: 2020-07-21 20:59:00
#
# Data Manager Class
#
import copy
import random
import os
import json
import pickle

from tqdm import tqdm
from numpy.lib.function_base import iterable
from ..utils import MyEncoder


def _DEFAULT_KEY_CALLBACK(rec):
    return rec['info']['uuid']

class DataManagerBase(object):
    def __init__(self, record_list, class_list):
        assert isinstance(record_list, (list, tuple))
        self.record_list = record_list
        self.class_list = class_list

    def filter(self, condition_callback, verbose=False):
        """
        Filter data by the given condition
        Param:
            condition_callback: a callback function that returns a boolean value according to attributes
            verbose: if set True, then print information about the operation
        Return:
            a DataManager instance, whose record_list satisfies the given condition
        """
        try:
            if verbose:
                record_list = [copy.deepcopy(x) for x in tqdm(self.iterator(), total=len(self)) if condition_callback(x)]
            else:
                record_list = [copy.deepcopy(x) for x in self.iterator() if condition_callback(x)]
        except RuntimeError as e:
            print("condition_callback Error:", e)
        data = type(self)(record_list, copy.copy(self.class_list))
        if verbose:
            print("{} out of {} images are found meeting the given condition".format(len(record_list), len(self)))
        return data

    def __eq__(self, another_dataset):
        assert isinstance(another_dataset, type(self))
        # check class list
        if not all([x==y for x, y in zip(self.class_list, another_dataset.class_list)]):
            return False

        # check occurrence of uuids
        key_callback=lambda rec: rec['info']['uuid']
        uuid_dict1 = self.occurrence(key_callback)
        uuid_dict2 = another_dataset.occurrence(key_callback)
        if uuid_dict1 != uuid_dict2:
            return False

        return True


    def __len__(self):
        return len(self.record_list)

    def __getitem__(self, idx):
        return self.record_list[idx]

    def clone(self):
        return copy.deepcopy(self)

    def iterator(self, shuffle=False):
        random_ids = list(range(len(self)))
        if shuffle:
            random.shuffle(random_ids)
        for idx in random_ids:
            yield self.record_list[idx]

    def shuffle(self):
        """ return a shuffled iterator (record can be edit in place) """
        return self.iterator(shuffle=True)

    def record_shuffle(self):
        random.shuffle(self.record_list)

    def split(self, num_or_ratio, groupID_callback=None, random_seed=123):
        """
        Split data by the given condition
        Param:
            num_or_ratio: number (int) or ratio (float) of the first half of the splits (by groupID)
            groupID_callback: a callback function that returns a hashable object (groupID); if unset, split the dataset itemwise
        Return:
            two DataManager instances, the groupID computed from the record_list 
            of one differs from the values computed from the record_list of the other
        """
        # 1) put data in group (by their groupID)
        groups = dict()
        for ix, record in enumerate(self.iterator()):
            if groupID_callback == None:
                group_id = ix
            else:
                try:
                    group_id = groupID_callback(record)
                except RuntimeError as e:
                    print("groupID_callback Error:", e)
            if group_id not in groups:
                groups[group_id] = [record]
            else:
                groups[group_id].append(record)
        num_groups = len(groups)

        # 2) split data by groups
        if None: pass
        elif iterable(num_or_ratio):
            # support ratio split like split([7,2,1])
            ratios = list(num_or_ratio) + [0]
            datasets = list()
            data = self.clone()
            for idx in range(len(ratios)-1):
                r = ratios[idx] / sum(ratios[idx:])
                ds, data = data.split(r, groupID_callback, random_seed)
                datasets.append(ds)
            assert len(data) == 0
            return datasets
        elif isinstance(num_or_ratio, float):
            assert 0 <= num_or_ratio <= 1
            num_firsthalf = int(num_groups * num_or_ratio)
        elif isinstance(num_or_ratio, int):
            num_firsthalf = num_or_ratio

        assert num_firsthalf <= num_groups, "There're less than {} groups to split: {} < {}".format(num_firsthalf, num_groups, num_firsthalf)

        # split data
        groups = [groups[key] for key in sorted(groups.keys())]  # sort groups by key for repeatibility
        random.seed(random_seed)
        random.shuffle(groups)
        firsthalf, lasthalf = groups[:num_firsthalf], groups[num_firsthalf:]
        record_list1 = list()
        for grp in firsthalf:
            record_list1.extend([copy.deepcopy(x) for x in grp])
        record_list2 = list()
        for grp in lasthalf:
            record_list2.extend([copy.deepcopy(x) for x in grp])

        data1 = type(self)(record_list1, copy.copy(self.class_list))
        data2 = type(self)(record_list2, copy.copy(self.class_list))

        return data1, data2

    def union(self, another_dataset):
        """
        Fuse two datasets
        Param:
            another_dataset: a DataManager instance
        Return:
            a DataManager instance that fused self and another_dataset
        """
        assert isinstance(another_dataset, type(self))
        assert all([x==y for x, y in zip(self.class_list, another_dataset.class_list)])
        record_list_merged = [copy.deepcopy(x) for x in self.record_list+another_dataset.record_list]
        data_merged = type(self)(record_list_merged, copy.copy(self.class_list))
        return data_merged

    def merge(self, another_dataset):
        return self.union(another_dataset)

    def dump(self):
        """ Print all info of the dataset """
        import pprint
        for ix, record in enumerate(self.iterator()):
            print("\n[{}/{}] sample >>".format(ix+1, len(self)))
            pprint.pprint(record)
        print("\nClass Dict:")
        print(self.class_list)

    def extract_info(self, info_callback):
        """
        Extract infomation extracted by the info_callback function
        Param:
            info_callback: callback function that extracts info from record
        Return:
            a list of info items extracted by info_callback
        """
        return list(map(info_callback, self.iterator()))

    def occurrence(self, key_callback=_DEFAULT_KEY_CALLBACK):
        """
        Count the occurrence of record
        Param:
            key_callback: a callback function that returns a hashable object (ie. groupID) or a list of hashable objects
        Return:
            a dictionary: {key_callback(record1): occurrence1, ...}
        """
        occurrence = dict()
        for record in self.iterator():
            try:
                key = key_callback(record)
            except RuntimeError as e:
                print("key_callback Error:", e)
            # if the extract key is a list (or tuple)
            if isinstance(key, list) or isinstance(key, tuple):
                keys = key
            else:
                keys = [key]
            for key in keys:
                if not key in occurrence:
                    occurrence[key]  = 1
                else:
                    occurrence[key] += 1
        return occurrence

    def unique(self, key_callback=_DEFAULT_KEY_CALLBACK, verbose=False):
        """
        Remove duplicated record by its key (the removal choices are made randomly)
        Param:
            key_callback: a callback function that returns a hashable object (ie. groupID) or a list of hashable objects
            verbose: if set True, then print information about the operation
        Return:
            a new DataManager object, whose records have unique key returned by key_callback
        """
        class_list = copy.deepcopy(self.class_list)
        record_list = list()
        visited_key = set()
        for record in self.iterator():
            try:
                key = key_callback(record)
            except RuntimeError as e:
                print("key_callback Error:", e)
            if key not in visited_key:
                record_list.append(copy.deepcopy(record))
                visited_key.add(key)
        if verbose:
            print("{} out of {} duplicated records are found".format(len(self)-len(record_list), len(self)))
        return type(self)(record_list=record_list, class_list=class_list)

    def intersection(self, another_dataset, key_callback=_DEFAULT_KEY_CALLBACK, verbose=False):
        """
        Find the intersection set of two datasets, if two records share the same key, they are deemed as members of the intersection set
        Param:
            another_dataset: a DataManager instance
            key_callback: a callback function that returns a hashable object (ie. image uuid or image hash code)
            verbose: if set True, then print information about the operation
        Return:
            a DataManager instance that represents for the intersection of two datasets
        """
        assert isinstance(another_dataset, type(self))
        assert all([x==y for x, y in zip(self.class_list, another_dataset.class_list)])
        class_list = copy.deepcopy(self.class_list)
        intersection_keys = set(another_dataset.extract_info(info_callback=key_callback))
        record_list = list()
        for record in self.iterator():
            try:
                key = key_callback(record)
            except RuntimeError as e:
                print("key_callback Error:", e)
            if key in intersection_keys:
                record_list.append(copy.deepcopy(record))
        if verbose:
            print("{} out of {} overlapping records are found".format(len(record_list), len(self)))
        return type(self)(record_list=record_list, class_list=class_list)
        
    def difference(self, another_dataset, key_callback=_DEFAULT_KEY_CALLBACK, verbose=False):
        """
        Find the difference set of two datasets, whose members exist in self but not in another_dataset
        Param:
            another_dataset: a DataManager instance
            key_callback: a callback function that returns a hashable object (ie. image uuid or image hash code)
            verbose: if set True, then print information about the operation
        Return:
            a DataManager instance that represents for the difference set of two datasets
        """
        assert isinstance(another_dataset, type(self))
        assert all([x==y for x, y in zip(self.class_list, another_dataset.class_list)])
        class_list = copy.deepcopy(self.class_list)
        intersection_keys = set(another_dataset.extract_info(info_callback=key_callback))
        record_list = list()
        for record in self.iterator():
            try:
                key = key_callback(record)
            except RuntimeError as e:
                print("key_callback Error:", e)
            if key not in intersection_keys:
                record_list.append(copy.deepcopy(record))
        if verbose:
            print("{} out of {} different records are found".format(len(record_list), len(self)))
        return type(self)(record_list=record_list, class_list=class_list)


    def data_statistics(self):
        raise NotImplementedError()


    def save_json(self, json_file):
        """
        Save dataset to json file
        Param:
            json_file: path to the json file where you want to save your dataset
        """
        json_dir = os.path.dirname(json_file)
        if json_dir:
            os.makedirs(json_dir, exist_ok=True)

        data_statistics = self.data_statistics()
        json_object = {
            'data_statistics': self.data_statistics(),
            'record': self.record_list,
            'class_list': self.class_list,
        }

        # save json object
        with open(json_file, 'w') as fid:
            json.dump(
                json_object,
                fid,
                indent = 4,
                cls=MyEncoder
            )


    def save_pickle(self, pickle_file):
        """
        Save dataset to pickle file
        Param:
            pickle_file: path to the pickle file where you want to save your dataset
        """
        pickle_dir = os.path.dirname(pickle_file)
        if pickle_dir:
            os.makedirs(pickle_dir, exist_ok=True)

        data_statistics = self.data_statistics()
        pickle_object = {
            'data_statistics': self.data_statistics(),
            'record': self.record_list,
            'class_list': self.class_list,
        }

        # save json object
        with open(pickle_file, 'wb') as fid:
            pickle.dump(pickle_object, fid)

    @classmethod
    def from_json(cls, json_file):
        """
        Load dataset from json file
        Param:
            json_file: path to the json file where you want to read. Its format has to satisfy: {'class_list': ..., 'record': ...}
        Return:
            a DataManager instance
        """
        assert os.path.exists(json_file), "{} doesn't exist".format(json_file)
        with open(json_file) as fid:
            obj = json.load(fid)

        assert 'record' in obj, "invald json file to load: {}".format(json_file)
        if 'class_list' in obj:
            class_list = obj['class_list']
        elif 'class_dict' in obj:
            class_list = obj['class_dict']
        else:
            raise RuntimeError("invald json file to load: {}".format(json_file))

        record_list = list()
        for rec in obj['record']:
            record_list.append(rec)
        dataset = cls(record_list=record_list, class_list=class_list)

        return dataset

    @classmethod
    def from_pickle(cls, pickle_file):
        """
        Load dataset from pickle file
        Param:
            pickle_file: path to the json file where you want to read. Its format has to satisfy: {'class_list': ..., 'record': ...}
        Return:
            a DataManager instance
        """
        assert os.path.exists(pickle_file), "{} doesn't exist".format(pickle_file)
        with open(pickle_file, 'rb') as fid:
            obj = pickle.load(fid)

        assert 'class_list' in obj and 'record' in obj, "invald json file to load: {}".format(pickle_file)
        class_list = obj['class_list']
        record_list = list()
        for rec in obj['record']:
            record_list.append(rec)
        dataset = cls(record_list=record_list, class_list=class_list)

        return dataset

    @classmethod
    def load(cls, filepath):
        ext = os.path.basename(filepath).split('.')[-1].lower()
        if ext in ('pickle', 'pkl'):
            return cls.from_pickle(filepath)
        else:
            return cls.from_json(filepath)

    def save(self, filepath, *args):
        ext = os.path.basename(filepath).split('.')[-1].lower()
        if ext in ('pickle', 'pkl'):
            return self.save_pickle(filepath, *args)
        else:
            return self.save_json(filepath, *args)

    def batch(self, batch_size, shuffle=False, random_seed=123):
        """
        Divide the dataset into a series of batches, randomly
        Param:
            batch_size: #records this method will return per iteration
        Return:
            an iterator of DataManager objects
        """
        assert isinstance(batch_size, int), "batch_size should be a natural number"
        if batch_size > len(self):
            batch_size = len(self)
        assert 0 < batch_size
        ds = self.clone()
        record_list = ds.record_list
        if shuffle:
            random.seed(random_seed)
            random.shuffle(record_list)
        for idx in range(0, len(record_list), batch_size):
            records = record_list[idx:idx+batch_size]
            yield type(self)(record_list=records, class_list=ds.class_list)

    def chunk(self, chunk_num, shuffle=False, random_seed=123):
        """
        Chunk the dataset (divide it into a given number of chunks), randomly
        Param:
            chunk_num: #chunks this method will return
        Return:
            an iterator of DataManager objects
        """
        assert isinstance(chunk_num, int), "chunk_num should be a natural number"
        assert 0 < chunk_num <= len(self)
        batch_size = len(self) // chunk_num
        if len(self) % chunk_num > 0:
            batch_size += 1
        return self.batch(batch_size, shuffle, random_seed)

    def zip(self, *args, key_callback=_DEFAULT_KEY_CALLBACK):
        """
        Iterate over multiple datasets
        Param:
            args: other datasets
            key_callback: a callback function that takes a record and return a 
                          hashable object which is used to align elements between
                          different datasets. If there're more than one records
                          sharing the same key, return them as a list during iteration
        Return:
            an iterator of record tuples
        Usage:
            for rec1, rec2, rec3 in data1.zip(data2, data3):
                ...
        """
        def _pack_records(dataset):
            rec_dict = dict()
            for rec in dataset:
                key = key_callback(rec)
                if key not in rec_dict:
                    rec_dict[key] = [rec]
                else:
                    rec_dict[key].append(rec)
            return rec_dict

        # prepare data
        rec_dict_list = [_pack_records(self)]
        for dataset in args:
            rec_dict = _pack_records(dataset)
            assert len(rec_dict) == len(rec_dict_list[0]), "Unmatched dataset under current key_callback"
            rec_dict_list.append(rec_dict)
        
        # loop over records that shares the same key among dataset
        for key in rec_dict_list[0].keys():
            ret = list()
            for dataset in rec_dict_list:
                recs = dataset[key]
                if len(recs) == 1:
                    ret.append(recs[0])
                else:
                    ret.append(recs)
            yield ret

    def __mul__(self, obj):
        if isinstance(obj, (int, float)):
            number = int(obj)
            assert number > 0
            new_record_list = self.record_list * number
            return (type(self))(new_record_list, self.class_list)
        else:
            assert isinstance(obj, type(self)), type(obj)
            another_dataset = obj
            return self.intersection(another_dataset)
    
    # overwrite operator '+'
    __add__ = union
    __or__  = union
    __and__ = intersection
    __sub__ = difference
    __truediv__ = difference



from .array_processing import *
from .box_processing import *
from .circle_processing import *
from .image_processing import *
from .math_processing import *
from .multi_processing import *
from .net_processing import *


import math
import numpy as np
from ..utils import cal_distance
from ..utils import is_iterable


def cartesian_to_polar(point_xy, center_xy, reverse_y=True):
    x, y = point_xy
    center_x, center_y = center_xy
    r = cal_distance(point_xy, center_xy)
    if r == 0:
        theta = 0
    elif r > 0:
        y_ = y - center_y
        if reverse_y:
            y_ = - y_
        x_ = x - center_x
        theta = math.asin(abs(y_) / max(1, abs(r)))
        if y_ >= 0 and x_ >= 0:
            pass
        elif y_ >= 0 and x_ < 0:
            theta = np.pi - theta
        elif y_ < 0 and x_ < 0:
            theta = np.pi + theta
        elif y_ < 0 and x_ >= 0:
            theta = 2 * np.pi - theta
        else:
            raise RuntimeError(f'bad y_ {y_} x_ {x_}')
    else:
        raise RuntimeError(f'bad distance {r}')
    return r, theta


def polar_to_cartesian(r, theta, center_xy, reverse_y=False):
    center_x, center_y = center_xy
    if reverse_y:
        y = center_y - r * math.sin(theta)
    else:
        y = center_y + r * math.sin(theta)
    x = center_x + r * math.cos(theta)

    point_xy = [x, y]
    return point_xy


class LogicOp:
    @staticmethod
    def EqualTo(value):
        def func(input):
            if input == value:
                return True
            else:
                return False
        return func

    @staticmethod
    def NotEqualTo(value):
        def func(input):
            if input != value:
                return True
            else:
                return False
        return func

    @staticmethod
    def LargerThan(value):
        def func(input):
            if input > value:
                return True
            else:
                return False
        return func

    @staticmethod
    def LessThan(value):
        def func(input):
            if input < value:
                return True
            else:
                return False
        return func

    @staticmethod
    def NotLessThan(value):
        def func(input):
            if input >= value:
                return True
            else:
                return False
        return func
    
    @staticmethod
    def NotLargerThan(value):
        def func(input):
            if input <= value:
                return True
            else:
                return False
        return func
    
    @staticmethod
    def In(value):
        assert is_iterable(value)
        def func(input):
            if input in value:
                return True
            else:
                return False
        return func

    @staticmethod
    def NotIn(value):
        assert is_iterable(value)
        def func(input):
            if input in value:
                return False
            else:
                return True
        return func
    
    @staticmethod
    def Is(value):
        def func(input):
            if input is value:
                return True
            else:
                return False
        return func
    
    @staticmethod
    def IsNot(value):
        def func(input):
            if input is not value:
                return True
            else:
                return False
        return func
    
    @staticmethod
    def IsTrue():
        def func(input):
            if input:
                return True
            else:
                return False
        return func
    
    @staticmethod
    def IsFalse():
        def func(input):
            if not input:
                return True
            else:
                return False
        return func





from .multi_label import ClassificationEvaluator, ClassificationMultiLabelEvaluator, MultiLabelEvaluator
from .multi_class import ClassificationMultiClassEvaluator, MultiClassEvaluator
from .detection import DetectionEvaluator


import sys
sys.path.append('.')

import pytest
import numpy as np
import re
import time
from lib.utils.utils import *

test_file = 'assets/everytime3p.jpg'

def test_get_list_from_list():
    test_data = list(range(20))
    res = get_list_from_list(test_data, lambda x: x if not x % 3 else None)
    assert res == [x  for x in range(20) if not x % 3 ]

    data = np.array(range(20))
    res = get_list_from_list(data, lambda x: x if not x % 3 else None)
    assert res == [x  for x in range(20) if not x % 3 ]

    data = {x : x for x in range(20)}
    res = get_list_from_list(data, lambda x: x if not x % 3 else None)
    assert res == [x  for x in range(20) if not x % 3 ]

def test_segment_intersection():
    seg_1 = (5, 19)
    seg_2 = (16, 24)
    res = segment_intersection(seg_1, seg_2)
    assert res == (3, (16, 19))

def test_concat_generator():
    from inspect import isgenerator
    def gen1():
        for index in range(10):
            yield index
    def gen2():
        for index in range(30):
            yield index % 10
    gen3 = concat_generator(gen1(), gen1(), gen1())
    assert isgenerator(gen3)
    for var1, var2 in zip(gen3, gen2()):
        assert var1 == var2

def test_get_mac_address():
    mac = get_mac_address()
    assert re.match("([A-Z0-9]{2}:){5}[A-Z0-9]{2}", mac) is not None

def test_cal_distance():
    vector_1, vector_2 = [1,2,3,4], [2,3,4,5]
    assert cal_distance(vector_1, vector_2) == 2

def test_get_file_size_M():
    res = get_file_size_M(test_file)
    assert 0.008 < res < 0.01

def test_unify_data_to_python_type():
    data = np.ones([2, 2]), 234, 152.5, True, Path('test')
    res = unify_data_to_python_type(data)
    assert [[[1.0, 1.0], [1.0, 1.0]], 234, 152.5, True, 'test'] == res

def test_timer_vvd():
    def test():
        sum_value = 0
        for index in range(100001):
            sum_value += index
        return sum_value
    res = timer_vvd(test)()
    assert res == (1+100000)*50000

def test_file_lines():
    data = ['abc', 'def']
    temp_data_path = 'assets/temp.txt'
    file_write_lines(data, temp_data_path, overwrite=True)
    lines = file_read_lines(temp_data_path)
    assert lines == data
    remove_file(temp_data_path)
    assert not OS_exists(temp_data_path)

def test_pickle():
    data = ['abc', 'def']
    temp_data_path = 'assets/temp.pickle'
    pickle_save(data, temp_data_path)
    load_data = pickle_load(temp_data_path)
    assert load_data == data
    remove_file(temp_data_path)
    assert not OS_exists(temp_data_path)

def test_find_sub_string():
    test_str = 'abc_0abc_1abc_2abc_3abc_4abc_5abc_6'
    res = find_sub_string(test_str, 'abc_', 5)
    assert test_str[res+4] == '5'

def test_str_connection():
    sub_1,sub_2, sub3 = 'asdf', 'asodkfj', 'askljdg'
    con = '&&&'
    res = str_connection(sub_1,sub_2, sub3, connect_char=con)
    assert res == 'asdf&&&asodkfj&&&askljdg'

def test_get_main_file_name():
    test_file_name = __file__
    res = get_main_file_name(test_file_name)
    assert res == 'test_utils'

def test_strong_printing():
    try:
        strong_printing('test')
        assert True
    except:
        assert False

def test_current_system():
    try:
        res = current_system()
        assert isinstance(res, str)
    except:
        assert False

def test_current_split_char():
    try:
        res = current_split_char()
        assert res in ['/', '\\']
    except:
        assert False

def test_save_file_path_check():
    test_file = __file__
    res = save_file_path_check(test_file)
    assert len(res) - len(test_file) == 20

def test_has_chinese_char():
    assert not has_chinese_char('asdlfkj')
    assert has_chinese_char('测试')

def test_encode_chinese_to_unicode():
    assert 'abc #U6d4b#U8bd5' == encode_chinese_to_unicode("abc 测试")
    assert 'abc ' == encode_chinese_to_unicode("abc 测试", remove=True)

def test_create_uuid():
    uuid = create_uuid()
    assert len(uuid) == 32

def test_xor():
    assert xor(True, False)
    assert not xor(True, True)
    assert not xor(False, False)

def test_get_file_hash_code():
    hash_code = get_file_hash_code(test_file)
    assert 'b5fc9a14763c0c83939bacb7ae3a664c' == hash_code

def test_get_date_str():
    string = get_date_str()
    assert re.match('[0-9]{4}-[0-9]{1,2}-[0-9]{1,2}', string) is not None

def test_isdir():
    parent_dir = Path(__file__).parent
    assert isdir(parent_dir)

def test_uniform_split_char():
    test_str = '\\asdfasdf\\asdfasdf/asdfasdf/asdfasdf/asdfasdf\\ewreg/asdfgsdfg'
    res1 = uniform_split_char(test_str)
    res2 = uniform_split_char(test_str, '\\')
    assert res1 == '/asdfasdf/asdfasdf/asdfasdf/asdfasdf/asdfasdf/ewreg/asdfgsdfg'
    assert res2 == '\\asdfasdf\\asdfasdf\\asdfasdf\\asdfasdf\\asdfasdf\\ewreg\\asdfgsdfg'
    pass

def test_dir_check():
    test_dir1 = Path(__file__).parent
    assert dir_check(test_dir1)
    test_dir2 = 'assets/test_dir/test_dir/test_dir'
    remove_dir('assets/test_dir')
    assert not dir_check(test_dir2)
    assert isdir(test_dir2)
    remove_dir('assets/test_dir')
    assert not isdir('assets/test_dir')

def test_time_reduce():
    assert 120 == time_reduce(1,2,3,4,5)

def test_get_function_name():
    res = get_function_name()
    assert res == 'test_get_function_name'

def test_draw_RB_map():
    y_true = [1] * 10 + [0] * 10
    y_pred = [res / 10 for res in range(10)] * 2
    file_path = 'assets/test.jpg'
    draw_RG_map(y_true, y_pred, map_save_path=file_path)
    assert OS_exists(file_path)
    remove_file(file_path)

def test_histgrom():
    value, bins = histgrom(range(100), show=False)
    assert all(value == 1)
    assert min(bins) == 0
    assert max(bins) == 99

def test_is_path_obj():
    path1 = Path('test')
    path2 = Path2('test')
    path3 = dict()
    path4 = 'str'
    assert is_path_obj(path1)
    assert is_path_obj(path2)
    assert not is_path_obj(path3)
    assert not is_path_obj(path4)

def test_time_stamp():
    stamp = time_stamp()
    assert re.match('[0-9]{4}-[0-9]{2}-[0-9]{2}_[0-9]{2}-[0-9]{2}-[0-9]{2}', stamp) is not None

def test_smart_copy():
    source_path = test_file
    new_path = change_file_name_for_path(source_path, 'test.asldkfj', '')
    smart_copy(source_path, new_path)
    assert get_file_hash_code(source_path) == get_file_hash_code(new_path)
    remove_file(new_path)

def test_json():
    data = [{'setasdf':'asdfasdf', '9234': {'afdasdf':'asdfasdf'}}, [1234, 'asdfasdf'], True]
    json_path = 'assets/test.json'
    json_save(data, json_path, overwrite=True)
    load_data = json_load(json_path)
    assert load_data == data
    remove_file(json_path)

def test_glob_recursively():
    json_path_list = glob_recursively('assets', 'json')
    assert len(json_path_list) > 0
    for json_path in json_path_list:
        assert OS_exists(json_path)

def test_is_integer():
    assert is_integer(3)
    assert is_integer(-3)
    assert is_integer(-0)
    assert is_integer(np.array(4))
    assert is_integer(np.array(4, dtype='int32'))
    assert is_integer(np.array([3, 4]))
    assert not is_integer(3.5)
    assert not is_integer(np.array(3.5))
    assert not is_integer(np.array([3, 3.5]))

def test_if_float():
    assert is_float(3.5)
    assert is_float(np.array(3.5))
    assert not is_float(3)
    assert not is_float(np.array(3))

def test_is_number():
    assert is_number(3)
    assert is_number(.3)
    assert not is_number('a')

def test_is_bool():
    assert is_bool(3 > 0)
    assert is_bool(np.array([3,4]) > 0)
    assert not is_bool(3)
    assert is_bool(True)
    assert is_bool(np.array(True))

def test_whether_divisible_by():
    assert whether_divisible_by(99, 3)
    assert not whether_divisible_by(97, 3)

def test_vvd_round():
    assert vvd_round(3.5) == 4
    assert vvd_round(3.6) == 4
    assert vvd_round(-3.6) == -4
    assert vvd_round([-3.6, 2.1]) == [-4, 2]

def test_vvd_ceil():
    assert vvd_ceil(3.6) == 4
    assert vvd_ceil(-3.6) == -3
    assert vvd_ceil([-3.6, 2.1]) == [-3, 3]

def vvd_floor():
    assert vvd_floor(3.6) == 3
    assert vvd_floor(-3.6) == -4
    assert vvd_floor([-3.6, 2.1]) == [-4, 2]

def test_get_gpu_str_as_you_wish():
    gpu_str, index_list = get_gpu_str_as_you_wish(1)
    assert isinstance(gpu_str, str)
    assert isinstance(index_list, list)
    gpu_str, index_list = get_gpu_str_as_you_wish(100)
    assert isinstance(index_list, list)
    res = get_single_gpu_id()
    assert isinstance(res, int)

def test_get_dir_file_list():
    dir_list, file_list = get_dir_file_list('.')
    for dir in dir_list:
        assert isdir(dir)
    for file in file_list:
        assert not isdir(file)
        assert OS_exists(file)
    all_dir_list, all_file_list = get_dir_file_list('.', recursive=True)
    assert len(all_dir_list) > len(dir_list)
    assert len(all_file_list) > len(file_list)

def test_get_segments():
    data = [0] * 3 + [1] * 30 + [0] * 13 + [1] * 3 + [0] * 3 + [1] * 3
    res = get_segments(data)
    assert res == [[3, 32], [46, 48], [52, 54]]

def test_try_exc_handler():
    def error():
        raise RuntimeError
    def exc_func():
        return 'abc'
    def ok():
        return True
    try:
        assert try_exc_handler(ok, error, True) == True
        assert True
    except:
        assert False

    try:
        assert try_exc_handler(error, ok) != True
        assert True
    except:
        assert False

    try:
        try_exc_handler(error, ok, True)
        assert False
    except:
        assert True
    
    try:
        try_exc_handler(error, error)
        assert True
    except:
        assert False
    assert try_exc_handler(exc_func, error) == 'abc'

def test_class_timer():
    class test:
        @staticmethod
        def test():
            return 'aaa'
    
    timmer = class_timer(test)
    assert timmer.test() == 'aaa'

def test_remove_small_components():
    mask = np.zeros([20, 20])
    mask[:3, :3] = 1
    mask[8:16, 8:16] = 1
    res = remove_small_components(mask, 10)
    assert np.sum(res) == 64

def test_encode_path():
    test_path = __file__
    assert 'test_utils@test_utils.py' == encode_path(test_path, 2)
    assert 'tests@test_utils@test_utils.py' == encode_path(test_path, 99, 'tests')
    assert 'tests&&&test_utils&&&test_utils.py' == encode_path(test_path, 99, 'tests', sep_char='&&&')
    assert 'abc/asdf/asdfasdf/asdf/tests-&test_utils-&test_utils.py' == encode_path(test_path, 99, 'tests', sep_char='-&', root_path='abc\\asdf\\asdfasdf/asdf')
    assert 'abc/asdf/asdfasdf/asdf/tests-&test_utils-&test_utils.ijiijijij' == encode_path(test_path, 99, 'tests', sep_char='-&', root_path='abc\\asdf\\asdfasdf/asdf', with_suffix='ijiijijij')
    assert 'tests-&test_utils-&test_utils' == encode_path(test_path, 99, 'tests', sep_char='-&', root_path='abc\\asdf\\asdfasdf/asdf', with_suffix='ijiijijij', stem=True)

def test_extract_file_to():
    root_dir = 'assets'
    target_dir = 'test_dir'
    extract_file_to(root_dir, 'jpg', target_dir)
    jpg_paths = glob_recursively(target_dir, 'jpg')
    assert jpg_paths
    remove_dir(target_dir)
    assert not OS_exists(target_dir)

def test_link():
    symlink('assets', 'test_dir')
    
    tar_temp_path = 'assets/test.json'
    assert hardlink('test_dir/mc.json', tar_temp_path)
    assert OS_exists(tar_temp_path)
    assert not islink(tar_temp_path)
    remove_file(tar_temp_path)
    assert not OS_exists(tar_temp_path)

    cwd = get_cwd()
    assert symlink(OS_join(cwd, 'assets/mc.json'), tar_temp_path)
    assert islink(tar_temp_path)
    data = json_load(tar_temp_path)
    assert OS_exists(tar_temp_path)
    remove_file(tar_temp_path)
    assert not OS_exists(tar_temp_path)
    remove_dir('test_dir')
    assert not OS_exists('test_dir')

def test_ListOrderedDict():
    obj = ListOrderedDict()
    obj.append('aa', 'aa')
    obj.append('ac', 'ac')
    obj.append('ad', 'ad')
    try:
        obj.append(3, 'ad')
        assert False
    except:
        assert True
    
    obj['bb'] = 1234
    obj['342'] = 'opq'
    
    assert obj[0] == 'aa'
    assert obj[1] == 'ac'
    assert obj[2] == 'ad'
    assert obj[3] == 1234
    assert obj[4] == 'opq'

def test_is_valid_identifier():
    assert not is_valid_identifier(123)
    assert not is_valid_identifier('123')
    assert not is_valid_identifier('')
    assert not is_valid_identifier('asdf?')
    assert is_valid_identifier('O')
    assert is_valid_identifier('_asdfa234')
    assert is_valid_identifier('iaslkdjflkajs')

def test_chinese_str_to_pinyin():
    assert 'asdf' == chinese_str_to_pinyin('asdf')
    assert 'as_zhe_shi_ge_ _sha_a_df' == chinese_str_to_pinyin('as这是个 啥啊df')
    assert 'astttzhetttshitttgettt tttshatttatttdf' == chinese_str_to_pinyin('as这是个 啥啊df', 'ttt')

def test_remove_chinese_for_file_names():
    try:
        remove_chinese_for_file_names('assets')
        assert True
    except:
        assert False

def test_get_file_time():
    res = get_file_time(test_file)
    for key, value in res.items():
        assert isinstance(value, time.struct_time)

def test_gen_iter():
    def gen():
        for index in range(20):
            yield index
    iter = list(range(10))
    assert is_generator(gen())
    assert is_generator(concat_generator(gen(), gen(), iter))
    assert is_iterable(iter)
    assert is_iterable(gen())
    assert not is_generator(iter)


def test_random_string():
    res_str = random_string(20)
    assert isinstance(res_str, str)
    assert len(res_str) == 20

def test_display_class_property():
    class Test:
        a = 3
        b = 'asdf'
    property_dict = display_class_property(Test)
    assert property_dict == {'a': 3, 'b': 'asdf'}


if __name__ == '__main__':
    pass
    # pytest.main()


from .data import DataManager
from .evaluator import ClassificationEvaluator, ClassificationMultiLabelEvaluator, MultiLabelEvaluator
from .evaluator import ClassificationMultiClassEvaluator, MultiClassEvaluator
from .evaluator import DetectionEvaluator
from .utils import *
from .labelme import *
from .processing import *



import sys
import setuptools
from setuptools import find_packages, setup
from numpy.distutils.core import setup
from numpy.distutils.misc_util import Configuration
from os.path import join, dirname, realpath

str_version = '2.1.1'



def configuration(parent_package='', top_path=''):
    # this will automatically build the scattering extensions, using the
    # setup.py files located in their subdirectories
    config = Configuration(None, parent_package, top_path)

    pkglist = setuptools.find_packages()
    for i in pkglist:
        config.add_subpackage(i)
    config.add_data_files(join('mtutils', 'assets', '*.json'))
    config.add_data_files(join('mtutils', 'assets', '*.jpg'))

    return config


if __name__ == '__main__':
    pass
    setup(
        configuration=configuration,
        name='mtutils',
        version=str_version,
        description='Commonly used function library by MT',
        url='https://github.com/zywvvd/utils_vvd',
        author='zywvvd',
        author_email='zywvvd@mail.ustc.edu.cn',
        license='MIT',
        packages=['mtutils'],
        zip_safe=False,
        install_requires= ['numba', 'func_timeout', 'pypinyin', 'opencv-python', 'sklearn'],
        python_requires='>=3')


# -*- coding: utf-8 -*-
# @Author: Zhang Yiwei
# @Date:   2020-07-18 02:40:35
# @Last Modified by:   Zhang Yiwei
# @Last Modified time: 2020-08-18 15:51:06
#
# vvd Tool functions
#
import inspect
import json
import time
import sys
import cv2 as cv
import logging
import os
import random
import platform
import hashlib
import pickle
import uuid
import math
import shutil
import string

import numpy as np
import PIL.Image as Image
import matplotlib.pyplot as plt

# 整合常用os操作
import os.path as osp
from os.path import basename as OS_basename
from os.path import join as OS_join
from os.path import exists as OS_exists
from os.path import isfile as OS_isfile
from os.path import isdir as OS_isdir
from os.path import dirname as OS_dirname

from numpy.lib.function_base import iterable
from pathlib2 import Path as Path2
from pathlib import Path
from glob import glob
from tqdm import tqdm
from collections import OrderedDict
from functools import wraps
from functools import reduce
from func_timeout import func_set_timeout, FunctionTimedOut

isfile = OS_isfile 
exists = OS_exists

def get_list_from_list(data_list, call_back, absolutely=False):
    """[make a list through a input list,
    while the output list will collect the output of a function dealing with every item of the input list]

    Args:
        data_list ([list/np.ndarray]): [original input list]
        call_back ([function]): [a call back function to do sth with every item of input list]
        absolutely([bool]): add result of call_back function to output_list whatever it is
    Returns:
        output_list[list]: [collection of output of call_back function]
    """
    output_list = list()
    if isinstance(data_list, np.ndarray):
        data_list = data_list.tolist()
    if isinstance(data_list, list):
        for data in data_list:
            res = call_back(data)
            if res is not None or absolutely:
                output_list.append(res)
    elif isinstance(data_list, dict):
        for key, data in data_list.items():
            res = call_back(data)
            if res is not None or absolutely:
                output_list.append(res)
    else:
        raise RuntimeError('input should be list or dict')
    return output_list


def segment_intersection(seg_1, seg_2):
    """ get intersection area from two segment

    Args:
        seg_1: two values which indicate start1 and end1
        seg_2 : two values which indicate start2 and end2

    Returns:
        value: length of intersection area
        tuple: (min, max)
    """
    assert len(seg_1) == len(seg_2) == 2
    min_value = max(min(seg_1), min(seg_2))
    max_value = min(max(seg_1), max(seg_2))
    inter_dis = max(0, max_value - min_value)
    return inter_dis, (min_value, max_value)


def concat_generator(*iterables):
    """ merge iterables to one generator
    Args: 
        iterables: Several unnamed parameters. All parameters are iterable variables
    
    Returns:
        Returns a generator that integrates all iterators
    """
    for it in iterables:
        for element in it:
            yield element


def get_mac_address():
    """get mac address

    Returns:
        str: mac address
    """
    mac = uuid.UUID(int = uuid.getnode()).hex[-12:].upper()
    return ":".join([mac[e: e+2] for e in range(0, 11, 2)])


def cal_distance(vector_1, vector_2, axis=None):
    """Calculate the Euclidean distance between the features 

    Args:
        vector_1: feature 1
        vector_2: feature 2
        axis: axis in numpy

    Returns:
        distance
    """
    assert len(vector_1) == len(vector_2) or len(vector_2) == 1
    distance = (np.sum((np.array(vector_1) - np.array(vector_2)) ** 2, axis=axis)) ** 0.5
    return distance


def get_file_size_M(file_path):
    """[get file size]

    Args:
        file_path ([str/Path]): [path to file]

    Returns:
        [float]: [size of file by M]
    """
    return os.path.getsize(str(file_path)) / 1024 / 1024


def unify_data_to_python_type(data):
    """[transfer numpy data to python type]

    Args:
        data ([dict/list]): [data to transfer]

    Returns:
        [type of input data]: [numpy data will be transfered to python type]
    """
    return json.loads(json.dumps(data, cls=MyEncoder))


def timer_vvd(func):
    """
    a timer for func
    you could add a @timer_vvd ahead of the fun need to be timed
    Args:
        func (function): a function to be timed

    Outputs:
        time message: a message which tells you how much time the func spent will be printed
    """
    is_static_method = False
    try :
        func_name = func.__name__
    except Exception as e:
        func_name = func.__func__.__name__
        func = func.__func__
        is_static_method = True

    @wraps(func)
    def wrapper(*args, **kwargs):
        if is_static_method:
            args = args[1:]

        start_time = time.time()
        res = func(*args, **kwargs)
        end_time = time.time()
        print('func: {_funcname_} runing: {_time_}s'.format(_funcname_=func_name, _time_=format(end_time - start_time, '.6f')))
        return res
    return wrapper


def file_rename(srcFile, dstFile):
    """rename file

    Args:
        srcFile (path): path to source file
        dstFile (path): path to target file
    """
    srcFile = str(srcFile)
    dstFile = str(dstFile)
    assert OS_exists(srcFile), f"srcFile {srcFile} not found"
    try:
        os.rename(srcFile, dstFile)
    except Exception as e:
        print(e)
        print('rename file fail\r\n')

def file_read_lines(file_path, encoding='utf8'):
    if not OS_exists(file_path):
        print("file {} not found, None will be return".format(file_path))
        return None

    with open(file_path, "r", encoding=encoding) as f:
        lines = f.readlines()
        for index, line in enumerate(lines):
            lines[index] = line.strip('\n')
        return lines


def file_write_lines(line_list, file_path, overwrite=False, verbose=False):
    dir_check(OS_dirname(file_path))
    file_path = save_file_path_check(file_path, overwrite, verbose)
    with open(file_path, 'w', encoding='utf8') as f:
        f.writelines('\n'.join(line_list))


def pickle_save(object, save_path, overwrite=False, verbose=False):
    """
    将object保存为pickle文件到save_path中
    """
    save_path = save_file_path_check(save_path, overwrite, verbose)
    with open(save_path, 'wb') as fp:
        pickle.dump(object, fp)


def pickle_load(load_path):
    """
    从load_path中读取object
    """
    assert isinstance(load_path, str) or is_path_obj(load_path)
    if isinstance(load_path, str):
        load_path = load_path.replace('\\', '/')
    with open(load_path, 'rb') as fp:
        return pickle.load(fp)


def find_sub_string(string, substring, times):
    """
    find the char position of the substring in string for times-th comes up
    """
    current = 0
    for _ in range(1, times+1):
        current = string.find(substring, current+1)
        if current == -1:
            return -1

    return current


def str_connection(*str_args, connect_char='_'):
    """
    connect strings in the list with underline
    """
    string = str.join(connect_char, str_args)
    return string


def get_main_file_name(string):
    """
    return file name without extension
    """
    assert isinstance(string, str) or is_path_obj(string)
    if is_path_obj(string):
        string = str(string)
    return os.path.splitext(os.path.basename(string))[0]


def strong_printing(*str_args):
    """
    print string strongly
    """
    assert isinstance(str_args, tuple)
    string = str_connection(*str_args, connect_char=' ')
    print()
    boudary_size = int(max(40, len(string)*1.4))
    split_string = boudary_size*'#'
    print(split_string)
    space_size = (boudary_size - len(string))//2
    print(space_size*' '+string.upper())
    print(split_string)
    print()


def current_system():
    """
    返回当前操作系统名称字符串
    """
    return platform.system()


def current_split_char():
    """
    返回当前操作系统的路径分隔符
    """
    if current_system() == 'Windows':
        return '\\'
    elif current_system() == 'Linux':
        return '/'
    else:
        return '/'


def save_file_path_check(save_file_path, overwrite=False, verbose=False):
    """
    检查要保存的文件路径
    - 如果文件已经存在 : 在文件名与扩展名之间加入当前时间作为后缀 避免覆盖之前的文件并给出提示
    - 如文件不存在 : 检查文件所在的文件夹目录
    返回检查后的文件路径
    """
    if is_path_obj(save_file_path):
        save_file_path = str(save_file_path)

    assert isinstance(save_file_path, str)
    if OS_exists(save_file_path):
        if overwrite:
            checked_save_file_path = save_file_path
            if verbose:
                print("file path {} already exists, the file will be overwrite.".format(save_file_path))
        else:
            main_file_name = get_main_file_name(save_file_path)
            new_base_name = OS_basename(save_file_path).replace(main_file_name, str_connection(main_file_name, time_stamp()))
            checked_save_file_path = OS_join(OS_dirname(save_file_path), new_base_name)
            if verbose:
                print("file path {} already exists, the file will be saved as {} instead.".format(save_file_path, checked_save_file_path))
    else:
        dir_check(str(Path(save_file_path).parent), verbose)
        assert OS_basename(save_file_path) != ''
        checked_save_file_path = save_file_path
    return checked_save_file_path


def has_chinese_char(input_str):
    """check if unASCII char in a string

    Args:
        input_str (str): test str

    Returns:
        True: there is at least one unACSII char in the input_str
        False: no unACSII char in the input_str
    """
    for char in input_str:
        if ord(char) > 255:
            return True
    return False


def encode_chinese_to_unicode(input_string, remove=False):
    '''
    将中文转换为 unicode #Uxxxx 形式
    Args:
        input_string: input string
        remove: if remove Chinese chars
    Returns:
        result string
    '''
    unicode_string = ''
    for char in input_string:
        if ord(char) > 255:
            if remove:
                continue
            char = "%%U%04x" % ord(char)
        unicode_string += char
    unicode_string = unicode_string.replace('%', '#')
    return unicode_string

def create_uuid():
    """ create a uuid (universally unique ID) with length 32"""
    md5_hash = hashlib.md5(uuid.uuid1().bytes)
    return md5_hash.hexdigest()

def xor(A, B):
    if (A and not B) or (B and not A):
        return True
    else:
        return False

def get_file_hash_code(file):
    """
    获取文件hash值
    """
    assert os.path.exists(file)
    md5_hash = hashlib.md5()
    with open(file, "rb") as fid:
        md5_hash.update(fid.read())
        digest = md5_hash.hexdigest()
    return digest


class MyEncoder(json.JSONEncoder):
    """
    自定义序列化方法, 解决 TypeError - Object of type xxx is not JSON serializable 错误
    使用方法:
    在json.dump时加入到cls中即可, 例如:
    json.dumps(data, cls=MyEncoder) 
    """

    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        elif isinstance(obj, np.bool_):
            return bool(obj)
        elif isinstance(obj, Path) or isinstance(obj, Path2):
            return str(obj)
        else:
            return super(MyEncoder, self).default(obj)


class Sys_Logger(object):
    '''
    修改系统输出流
    '''

    def __init__(self, fileN="Default.log"):

        self.terminal = sys.stdout
        if OS_exists(fileN):
            self.log = open(fileN, "a")
        else:
            self.log = open(fileN, "w")

    def write(self, message):

        self.terminal.write(message)
        self.log.write(message)

    def flush(self):
        self.terminal.flush()


class Loger_printer():
    """
    日志打印类
    会在控制台与日志同时打印信息    
    """

    def __init__(self, logger):
        self.logger = logger
        self.logger_dict = {
            logging.DEBUG: self.logger.debug,
            logging.INFO: self.logger.info,
            logging.WARNING: self.logger.warning,
            logging.ERROR: self.logger.error,
            logging.CRITICAL: self.logger.critical
        }
        self.closed = False

    def get_logger_func(self, level):
        if level in self.logger_dict:
            return self.logger_dict[level]
        else:
            raise RuntimeError(f"unknown level {level}, {self.logger_dict}")

    def vvd_logging(self, *message, level=logging.INFO, close_logger=False):
        if not self.closed:
            log_func = self.get_logger_func(level)
            if message is not None:
                for message_str in message:
                    print(message_str)
                    log_func(message_str)
            if close_logger:
                for handler in self.logger.handlers:
                    handler.close()
        else:
            print('logger has been closed')

    def vvd_logging_quiet(self, *message, level=logging.INFO, close_logger=False):
        if not self.closed:
            log_func = self.get_logger_func(level)
            if message is not None:
                for message_str in message:
                    log_func(message_str)
            if close_logger:
                for handler in self.logger.handlers:
                    handler.close()
        else:
            print('logger has been closed')

def get_date_str():
    str_time = time.localtime(time.time())
    year = str_time.tm_year
    mounth = str_time.tm_mon
    day = str_time.tm_mday

    date_str = f"{year}-{mounth}-{day}"
    return date_str


def log_init(log_path, quiet=False, level=logging.INFO):
    """
    initialize logging 
    save the logging object in `config.Parameters.Logging_Object`

    after this operation,
    we could save logs with simple orders such as `logging.debug('test debug')` `logging.info('test info')` 
    logging level : debug < info < warning <error < critical

    Loger_printer.vvd_logging('test')
    """
    log_path= str(log_path)
    dir_name = os.path.dirname(log_path)

    dir_check(dir_name)
    log_file_path = log_path

    if os.path.exists(log_file_path):
        # open log file as  mode of append
        open_type = 'a'
    else:
        # open log file as  mode of write
        open_type = 'w'

    # basicConfig 无法解决中文乱码的问题, 封存相关代码
    # 据说 Python 3.9 之后会加入basicConfig() 的encoding和errors关键字配置
    # logging.basicConfig(
    #     # 日志级别,logging.DEBUG,logging.ERROR
    #     level=level,
    #     # 日志格式: 时间、   日志信息
    #     format='%(asctime)s: %(message)s',
    #     # 打印日志的时间
    #     datefmt='%Y-%m-%d %H:%M:%S',
    #     # 日志文件存放的目录（目录必须存在）及日志文件名
    #     filename=log_file_path,
    #     # 打开日志文件的方式
    #     filemode=open_type
    # )
    # logging.StreamHandler()

    # create logger obj
    logger = logging.getLogger()
    # set log level
    logger.setLevel(level)
    # file handler 
    handler = logging.FileHandler(log_file_path, mode=open_type, encoding='utf-8')
    handler.setFormatter(logging.Formatter("%(asctime)s-%(name)s-%(levelname)s: %(message)s"))

    for old_handler in logger.handlers[::-1]:
        old_handler.stream.close()
        logger.removeHandler(old_handler)

    logger.addHandler(handler)

    if quiet:
        return Loger_printer(logger).vvd_logging_quiet
    else:
        return Loger_printer(logger).vvd_logging


def isdir(dir_path):
    """
    check if dir exists
    """
    dir_path = str(dir_path)
    if not os.path.isdir(dir_path):
        return False
    else:
        return True


def uniform_split_char(string, split_char='/'):
    """
    uniform the split char of a string
    """
    string = str(string)
    assert isinstance(string, str)
    return string.replace('\\', split_char).replace('/', split_char)


def dir_check(dir_path, verbose=False):
    """
    check if `dir_path` is a real directory path
    if dir not found, make one
    """

    dir_path = str(dir_path)
    if dir_path == '':
        return True
    assert isinstance(dir_path, str)
    dir_path = uniform_split_char(dir_path)
    if not os.path.isdir(dir_path):
        try:
            os.makedirs(dir_path)
            if verbose:
                print('dirs made: {}'.format(dir_path))
        except Exception as err:
            print(f'failed to make dir {dir_path}, error {err}')
        return False
    else:
        return True


def time_reduce(*data):
    """
    [计算输入数据的乘积]
    """
    data = list(data)
    return reduce(lambda x, y: x*y, data)


def get_function_name():
    '''获取正在运行函数(或方法)名称'''
    # print(sys._getframe().f_code.co_name)
    return inspect.stack()[1][3]


def draw_RG_map(y_true, y_pred, map_save_path=None):
    """draw Red Green map for a prediction
    the predicted data are sorted by score
    Correctly classified samples are marked in green
    The misclassified sample is marked in red

    Args:
        y_true (list or narray): ground truth
        y_pred (list or narray): prediction
        map_save_path (path): path to the output image
    """
    y_true = np.array(y_true)
    y_pred = np.array(y_pred)

    assert isinstance(y_pred, np.ndarray) and isinstance(y_true, np.ndarray)
    assert np.ndim(y_pred) == 1
    assert y_pred.shape == y_true.shape

    sorted_ids = np.argsort(y_pred+np.random.rand(y_pred.size)*1e-12)
    sorted_y_true = y_true[sorted_ids]
    ng_rank = np.where(sorted_y_true == 1)[0]
    ok_rank = np.where(sorted_y_true == 0)[0]

    plt.figure(figsize=(25, 3))
    plt.bar(ok_rank, 1, width=1, color='g')
    plt.bar(ng_rank, 1, width=int(ok_rank.size/ng_rank.size/5+1), color='r')
    plt.ylim([0, 1])
    plt.xlim([0, len(ok_rank)+len(ng_rank)])
    if map_save_path is not None:
        plt.savefig(map_save_path)
    else:
        plt.show()

        plt.figure(figsize=(25, 3))
        plt.hist(y_pred, bins=255)
        plt.title('ng_score distribution')
        plt.show()


def histgrom(*data, bin_num=100, show=True):
    data_list = list(data)
    value, bins = np.histogram(data_list, bins=bin_num)
    for data_item in data_list:
        plt.hist(data_item, bin_num)
    if show:
        plt.show()
    return value, bins


def data_show(data):
    '''
    show data in a chart
    '''
    plt.plot(data)
    plt.show()


def is_path_obj(path):
    """
    Determines whether the input path is a Path obj
    """
    if isinstance(path, Path) or isinstance(path, Path2):
        return True
    else:
        return False


def time_stamp():
    """
    返回当前时间戳字符串
    格式: 年-月-日_时-分-秒
    """
    return time.strftime("%Y-%m-%d_%H-%M-%S", time.localtime())

def get_suffix(path):
    """get suffix of a path

    Args:
        path (_type_): _description_
    """
    cur_path = Path(path)
    suffix = cur_path.suffix.replace('.', '')
    return suffix

def change_file_name_for_path(file_path, new_name=None, new_suffix=None):
    """change file name for a file path
    This operation is valid only for path, NOT for REAL FILE
    

    Args:
        file_path (path): a file path
        new_name (str): new file name
        new_suffix(str): new suffix, if None, old suffix will be kept

    Returns:
        new file path(str)
    """
    path = Path(str(file_path))
    parent = path.parent
    
    if new_name is None:
        new_file_stem = path.stem
    else:
        new_file_stem = new_name

    if new_suffix is not None:
        if new_suffix == '':
            new_file_suffix = ''
        else:
            if new_suffix[0] != '.':
                new_file_suffix = '.' + new_suffix
            else:
                new_file_suffix = new_suffix
    else:
        new_file_suffix = path.suffix

    new_file_name = new_file_stem + new_file_suffix
    new_path = parent / new_file_name
    new_path = uniform_split_char(str(new_path))
    return new_path

def smart_copy(source_file_path, target_path, verbose=False, remove_source_file=False, overwrite=False):
    """[复制文件从源到目标, 如果目标已经存在则跳过]]

    Args:
        source_file_path ([str]): [源文件路径]
        target_path ([str]): [目标文件夹/目标文件路径]
        verbose (bool, optional): [是否显示信息]. Defaults to False.
    """
    source_file_path = str(source_file_path).replace('\\', '/')
    target_path = str(target_path).replace('\\', '/')
    assert OS_exists(source_file_path)
    if OS_isdir(target_path):
        target_path = OS_join(target_path, OS_basename(source_file_path))
    exists = OS_exists(target_path)
    if exists and not overwrite:
        if verbose:
            print("{} already exists! skip copying.".format(target_path))
    else:
        dir_check(Path(target_path).parent)
        if exists and overwrite:
            if verbose:
                print("{} already exists! start overwriting.".format(target_path))
        if remove_source_file:
            shutil.move(source_file_path, target_path)
        else:
            shutil.copy(source_file_path, target_path)


def json_load(json_path, verbose=False):
    """
    读取json文件并返回内容字典
    """
    json_path = str(json_path)
    if isinstance(json_path, str):
        json_path = json_path.replace('\\', '/')
    try:
        assert OS_exists(json_path)
    except Exception as e:
        if verbose:
            print('file not found !', e)
    try:
        with open(json_path, 'r') as fp:
            return json.load(fp)
    except Exception as e:
        if verbose:
            print('simple json load failed, try utf-8', e)
    try:
        with open(json_path, 'r', encoding='utf-8') as fp:
            return json.load(fp)
    except Exception as e:
        if verbose:
            print('utf-8 json load failed, try gbk', e)
    try:
        with open(json_path, 'r', encoding='gbk') as fp:
            return json.load(fp)
    except Exception as e:
        if verbose:
            print('gbk json load failed!', e)        


def json_save(json_dict, json_path, overwrite=False, verbose=False):
    """
    将内容字典保存为json文件
    """
    json_path = str(json_path)
    json_path = save_file_path_check(json_path, overwrite, verbose)
    with open(json_path, 'w', encoding='utf-8') as fp:
        json.dump(json_dict, fp, ensure_ascii=False, sort_keys=False, indent=4, cls=MyEncoder)


def glob_recursively(path, extension, recursively=True):
    """
    在path 路径中递归查找所有扩展名为extension的文件, 返回完整路径名列表
    """
    path = str(path).replace('\\', '/')
    if recursively:
        return glob(OS_join(path, '**', '*.' + extension), recursive=True)
    else:
        return glob(OS_join(path, '*.' + extension), recursive=True)


def is_integer(num):
    """
    是否是整数, 返回bool结果
    """
    from numbers import Integral
    if isinstance(num, np.ndarray):
        return np.issubdtype(num.dtype, np.integer)
    else:
        return isinstance(num, Integral)


def is_float(num):
    """
    是否是浮点数, 返回bool结果
    """
    if isinstance(num, np.ndarray):
        return np.issubdtype(num.dtype, np.floating)
    else:
        return isinstance(num, float)


def is_number(num):
    """
    是否是数字, 此处暗指实数, 返回bool结果
    """
    return is_float(num) or is_integer(num)


def is_bool(data):
    """
    是否是 bool 值, 返回bool结果
    """
    if isinstance(data, np.ndarray):
        return np.issubdtype(data.dtype, np.bool_)
    return isinstance(data, bool)


def whether_divisible_by(to_be_divided, dividing):
    """
    to_be_divided 是否可以被 dividing 整除, 返回bool结果
    """
    assert is_integer(to_be_divided) and is_integer(dividing)
    if to_be_divided % dividing == 0:
        return True
    else:
        return False


def vvd_round(num):
    if iterable(num):
        return np.round(np.array(num)).astype('int32').tolist()
    return int(round(num))


def vvd_ceil(num):
    if iterable(num):
        return np.ceil(np.array(num)).astype('int32').tolist()
    return int(np.ceil(num))


def vvd_floor(num):
    if iterable(num):
        return np.floor(np.array(num)).astype('int32').tolist()
    return int(np.floor(num))


def get_gpu_str_as_you_wish(gpu_num_wanted, verbose=0):
    """[get empty gpu index str as needed]

    Args:
        gpu_num_wanted ([int]): [the num of gpu you want]

    Returns:
        [str]: [gpu str returned]
    """
    try:
        import pynvml
    except Exception as e:
        print('can not import pynvml.', e)
        print('please make sure pynvml is installed correctly.')
        print('a simple pip install nvidia-ml-py3 may help.')
        print('now a 0 will be return')
        return '0', [0]

    NUM_EXPAND = 1024 * 1024

    try:
        # 初始化工具
        pynvml.nvmlInit()
    except Exception as e:
        print('pynvml.nvmlInit failed:', e)
        print('now a 0 will be return')
        return '0', [0]

    # 驱动信息
    if verbose:
        print("GPU driver version: ", pynvml.nvmlSystemGetDriverVersion())

    # 获取Nvidia GPU块数
    gpuDeviceCount = pynvml.nvmlDeviceGetCount()
    returned_gpu_num = max(min(gpu_num_wanted, gpuDeviceCount), 0)

    gpu_index_and_free_memory_list = list()
    for index in range(gpuDeviceCount):
        handle = pynvml.nvmlDeviceGetHandleByIndex(index)
        info = pynvml.nvmlDeviceGetMemoryInfo(handle)
        gpu_memory_total = info.total / NUM_EXPAND
        gpu_memory_used = info.used / NUM_EXPAND
        gpu_index_and_free_memory_list.append([index, gpu_memory_total - gpu_memory_used])
        if verbose:
            if index == 0:
                device_name = pynvml.nvmlDeviceGetName(handle)
                print('GPU device name:', device_name)
            print(f'gpu {index}: total_memory: {gpu_memory_total} memory_left: {gpu_memory_total - gpu_memory_used}')

    gpu_index_and_free_memory_list.sort(key=lambda x: - x[1])

    gpu_index_picked_list = list()

    for i in range(returned_gpu_num):
        gpu_index_picked_list.append(gpu_index_and_free_memory_list[i][0])

    gpu_index_str = ','.join([str(index) for index in gpu_index_picked_list])
    if verbose:
        print(returned_gpu_num, 'gpu index will be return.')
        print(f'return gpu str: {gpu_index_str}')

    # 关闭工具
    pynvml.nvmlShutdown()

    return gpu_index_str, gpu_index_picked_list

def get_single_gpu_id(verbose=False):
    id_str, id_list = get_gpu_str_as_you_wish(1, verbose=verbose)
    gpu_id = id_list[0]
    return gpu_id

def get_dir_file_list(root_path, recursive=False):
    """[get dir and file list under root_path recursively]

    Args:
        root_path ([str]): [root dir to querry]
        recursive ([bool]): [whether walk recursively]

    Returns:
        dir_list [list]: [output dir list]
        file_list [list]: [output file list]
    """

    dir_list = list()
    file_list = list()

    root_path = str(root_path)

    for root, dirs, files in os.walk(root_path):
        file_list += get_list_from_list(files, lambda x: os.path.join(root, x))

        for dir in dirs:
            cur_dir_path = os.path.join(root, dir)
            dir_list.append(cur_dir_path)

        if not recursive:
            break
    dir_list = get_list_from_list(dir_list, uniform_split_char)
    file_list = get_list_from_list(file_list, uniform_split_char)
    return dir_list, file_list


def get_segments(data):
    """
    get segments for data (for moved safe data)
    """
    data = np.array(data)
    assert data.ndim == 1
    data = (data > 0).astype('int8')
    mark = data[:-1] - data[1:]
    start_pos = np.nonzero(mark == -1)[0].tolist()
    end_pos = np.nonzero(mark == 1)[0].tolist()
    if data[0] > 0:
        start_pos = [-1] + start_pos
    if data[-1] > 0:
        end_pos = end_pos + [len(mark)]
    assert len(start_pos) == len(end_pos)
    segments_list = [[x + 1, y] for x, y in zip(start_pos, end_pos)]
    return segments_list


def try_exc_handler(try_func, exc_func, developer_mode=False):
    except_result = try_result = None

    def exception_handler(e, exc_func):
        try:
            ori_exception_info = list(e.args)
            if len(ori_exception_info) == 0:
                ori_exception_info.append('')
            ori_exception_info[0] = ' ErrorMessage: ' + str(ori_exception_info[0])\
                + '\n CrashFile: ' + str(e.__traceback__.tb_next.tb_frame.f_globals['__file__'])\
                + '\n Line: ' + str(e.__traceback__.tb_next.tb_lineno)

            e.args = tuple(ori_exception_info)
            except_result = exc_func(e)
        except Exception as ee:
            print("!! exc_func failed.")
            print(f"!! error message: {str(ee)}")
            print("!! we can only return the previous exception info.")
            except_result = str(e)
        return except_result

    if developer_mode:
        try_result = try_func()
    else:
        try:
            try_result = try_func()
        except Exception as e:
            return exception_handler(e, exc_func)
        except FunctionTimedOut as e:
            return exception_handler(e, exc_func)

    return try_result


def class_timer(input_class):
    """
    get a new class through a Decorator
    new class will Time all funcs of origin class

    Args:
        input_class (class): class to be timed

    Returns:
        class: timer class
    """
    class Timer(input_class):
        def __getattribute__(self, name: str):

            func = super().__getattribute__(name)

            if str(type(func)) == "<class 'method'>":
                is_static_method = False
                try :
                    func_name = func.__name__
                except Exception as e:
                    func_name = func.__func__.__name__
                    func = func.__func__
                    is_static_method = True

                @wraps(func)
                def wrapper(*args, **kwargs):
                    if is_static_method:
                        args = args[1:]

                    start_time = time.time()
                    res = func(*args, **kwargs)
                    end_time = time.time()
                    print('func: {_funcname_} runing: {_time_}s'.format(_funcname_=func_name, _time_=format(end_time - start_time, '.6f')))
                    return res
                return wrapper
            return func
    return Timer


def remove_small_components(mask, area_thres):
    """ remove connected components from mask if their areas < area_thres """
    nccomps = cv.connectedComponentsWithStats(mask.astype(np.uint8))

    labels = nccomps[1]
    status = nccomps[2]

    for index, st in enumerate(status.tolist()):
        if st[0] == st[1] == 0:
            continue
        area = st[4]
        if area >= area_thres:
            labels[labels == index] = - index

    return labels < 0


def encode_path(path, seg_num = 4, key_word=None, sep_char='@', root_path=None, with_suffix=None, stem=False):
    """Change path to a string without slash

    Args:
        path (path): path to be encoded
        seg_num (int, optional): The num of subpath segments that need to be encoded. Defaults to 4.
        key_word (string, optional): if set(not None), the path will be truncated after the keyword. Defaults to None.
        sep_char (str, optional): a string who will concatenate subpaths. Defaults to '@'.
        root_path (str, optional): if set, a encoded file path under root_path will be return . Defaults to None.
        with_suffix (str, optional): if set, result path will use the input suffix. We both support suffixes start with . or not. Defaults to None.
        stem (bool, optional): if True, only stem of encoded path will be returned. Defaults to False.

    Returns:
        str: encoded path
    """
    path = uniform_split_char(str(path))
    if key_word is not None:
        pos = str(path).find(key_word)
        if pos == '-1':
            print('keyword encode path failed, use default function')
            key_word = None
        else:
            path = str(path)[pos:]
            tmp_path = Path(path)
            encoded_path = sep_char.join(tmp_path.parts)

    if key_word is None:
        tmp_path = Path(path)
        encoded_path = sep_char.join(tmp_path.parts[-seg_num:])

    if root_path is not None:
        encoded_path = str(Path(root_path) / encoded_path)

    if with_suffix is not None:
        with_suffix = str(with_suffix)
        if with_suffix[0] != '.':
            with_suffix = '.' + with_suffix
        encoded_path = str(Path(str(encoded_path)).with_suffix(with_suffix))

    if stem:
        encoded_path = Path(encoded_path).stem

    return uniform_split_char(encoded_path)


def extract_file_to(root_dir, suffix, target_dir, encode_level=3):
    """copy files with specified suffix in root dir to target dir
    new file name is encoded oirgin file path

    Args:
        root_dir (path): root dir
        suffix (str): suffix without .
        target_dir (path): target dir
        encode_level (int, optional): path encode level. Defaults to 3.
    """
    file_list = glob_recursively(root_dir, suffix)
    desc = f"Extracting {suffix} file from {root_dir} to {target_dir}:"
    for file_path in tqdm(file_list, desc=desc):
        file_name = encode_path(file_path, seg_num=encode_level)
        save_path = Path(target_dir) / file_name
        smart_copy(file_path, save_path)


def link(src, tar, hard=False):
    """build link for src file to tar file

    Args:
        src (path): source file
        tar (path): target file
        hard (bool): hard link if True, else symlink

    Returns:
        bool: if success return True, else False
    """

    if osp.exists(tar):
        return False
    else:
        if osp.exists(src):
            if hard:
                if not isdir(src):
                    os.link(src, tar)
                else:
                    return False
            else:
                os.symlink(src, tar)
        else:
            return False
    return True

def islink(path):
    """Test whether a path is a symbolic link"""
    return os.path.islink(path)

def get_cwd():
    """get current working directory"""
    return uniform_split_char(os.getcwd())

def hardlink(src, tar):
    """
    make hardlink for src and tar
    """
    return link(src, tar, True)

def symlink(src, tar):
    """
    make sumlink for src and tar
    """
    return link(src, tar, False)


class ListOrderedDict(OrderedDict):
    def __getitem__(self, key):
        if is_integer(key):
            key = list(self.keys())[key]
        elif isinstance(key, slice):
            slicedkeys = list(self.keys())[key]
            lod_obj = type(self)()
            for slicedkey in slicedkeys:
                lod_obj[slicedkey] = self[slicedkey]
            return lod_obj

        return super().__getitem__(key)

    def append(self, key, value):
        if is_integer(key):
            raise RuntimeError(f" integer key is not allowed in ListOrderedDict {key}.")
        assert key not in self, f"cannot add existed key. {key}"
        self[key] = value
    
    def __setitem__(self, key, value):
        'od.__setitem__(i, y) <==> od[i]=y'
        # Setting a new item creates a new link at the end of the linked list,
        # and the inherited dictionary is updated with the new key/value pair.
        if is_integer(key):
            raise RuntimeError(f" integer key is not allowed in ListOrderedDict {key}.")
        else:
            return super().__setitem__(key, value)

    def __delitem__(self, key):
        'od.__delitem__(y) <==> del od[y]'
        # Deleting an existing item uses self.__map to find the link which gets
        # removed by updating the links in the predecessor and successor nodes.
        if is_integer(key):
            key = list(self.keys())[key]

        return super().__delitem__(key)

    def __iter__(self): 
        index = 0
        while index < len(self):
            yield self[index]      # 使用 yield
            index = index + 1

    def move_to_end(self, key, last=True):
        '''Move an existing element to the end (or beginning if last is false).

        Raise KeyError if the element does not exist.
        '''
        if is_integer(key):
            key = list(self.keys())[key]

        return super().move_to_end(key, last)

    def __repr__(self):
        'od.__repr__() <==> repr(od)'
        if not self:
            return '%s()' % (self.__class__.__name__,)
        return '%s(%r)' % (self.__class__.__name__, list(self.items()))

    def pop(self, key=-1):
        '''od.pop(k[,d]) -> v, remove specified key and return the corresponding
        value.  If key is not found, d is returned if given, otherwise KeyError
        is raised.
        '''
        if is_integer(key):
            if len(list(self.keys())) == 0:
                print(f"pop a None from an empty ListOrderedDict.")
                return None
            key = list(self.keys())[key]
        return super().pop(key)

    def setdefault(self, key, default=None):
        '''Insert key with a value of default if key is not in the dictionary.

        Return the value for key if key is in the dictionary, else default.
        '''
        if is_integer(key):
            key = list(self.keys())[key]
        return super().setdefault(key, default)

def is_valid_identifier(input_str):
    """ Test whether a str is a identifier"""
    ascii = string.ascii_letters+'_'
    digits = string.digits
    if not isinstance(input_str, str):
        return False
    if len(input_str) == 0:
        return False
    if input_str[0] not in ascii:
        return False
    for char in input_str:
        if char not in ascii and char not in digits:
            return False
    return True


def chinese_str_to_pinyin(input_str: str, join_char='_') -> str:
    """change str with chinese char into pinyin chars

    Args:
        input_str (str): input string
        join_char (str, optional): string to split pinyin. Defaults to '_'.

    Returns:
        str: processed string
    """
    import pypinyin
    pinyin_str =  pypinyin.pinyin(input_str, style=pypinyin.NORMAL)
    pinyin_str_list = get_list_from_list(pinyin_str, lambda x: x[0])
    pinyin_res_str = join_char.join(pinyin_str_list)
    output_str = encode_chinese_to_unicode(pinyin_res_str, remove=True)
    return output_str


def remove_chinese_for_file_names(root_dir):
    """change all filenames of files in root_dir which have chinese char in file_name to pinyin names"""
    dir_list, file_list = get_dir_file_list(root_dir, recursive=True)
    file_name_changed = 0
    for file_path in tqdm(file_list, desc=f"changing file names: "):
        file_name = Path(file_path).name
        changed_name = chinese_str_to_pinyin(file_name)
        if file_name != changed_name:
            dst_path = Path(file_path).parent / changed_name
            file_rename(file_path, dst_path)
            file_name_changed += 1
    print(f"change {file_name_changed} file names.")
    
    dir_name_changed = 0
    for dir_path in tqdm(dir_list[::-1], desc=f"changing directory names: "):
        dir_name = Path(dir_path).name
        changed_name = chinese_str_to_pinyin(dir_name)
        if dir_name != changed_name:
            dst_path = Path(dir_path).parent / changed_name
            file_rename(dir_path, dst_path)
            dir_name_changed += 1
    print(f"change {dir_name_changed} dir names.")
    pass


def is_generator(obj):
    from inspect import isgenerator
    return isgenerator(obj)


def is_iterable(obj):
    from typing import Iterable
    return isinstance(obj, Iterable)


def get_file_time(file_path):
    """get file time record"""
    atime = os.path.getatime(file_path)  # 文件访问时间
    mtime = os.path.getmtime(file_path)  # 文件最近修改时间
    ctime = os.path.getctime(file_path)  # 文件的创建时间

    time_info = dict(
        access_time=time.localtime(atime),
        modify_time=time.localtime(mtime),
        create_time=time.localtime(ctime)
    )

    return time_info


def random_string(length, digits=True, letters=True, punctuation=True):
    """make random string via specific char sets

    Args:
        length (int): length of target random string
        digits (bool, optional): if True, random string will contain digits. Defaults to True.
        letters (bool, optional): if True, random string will contain letters. Defaults to True.
        punctuation (bool, optional): if True, random string will contain punctuations. Defaults to True.

    Returns:
        _type_: _description_
    """
    str_temp = ''
    if digits:
        str_temp = str_temp + string.digits
    if letters:
        str_temp = str_temp + string.ascii_letters
    if punctuation:
        str_temp = str_temp + string.punctuation
    return ''.join(random.choice(str_temp) for _ in range(length))


def remove_file(path):
    """remove a file or a symlink"""
    path = str(path)
    if OS_exists(path):
        if OS_isfile(path) or islink(path):
            os.remove(path)


def remove_dir(path):
    """remove a dir or a symlink"""
    path = str(path)
    if OS_exists(path):
        if islink(path):
            os.remove(path)
        else:
            shutil.rmtree(path)


def display_class_property(class_def, verbose=False):
    """Display Configuration values."""
    if verbose:
        print("\n")
        print("Configurations:")
    property_dict = dict()
    for property in dir(class_def):
        if not property.startswith("__") and not callable(getattr(class_def, property)):
            property_dict[property] = getattr(class_def, property)
            if verbose:
                print("{:50} {}".format(property, getattr(class_def, property)))
    if verbose:
        print("\n")
    return property_dict



if __name__ == '__main__':
    pass